@@ -41,7 +41,13 @@ __all__ = [
     'WADLError',
     ]
 
-from cStringIO import StringIO
+try:
+    # Make sure to try the Python 2 form first; we don't want to use
+    # io.StringIO in Python 2, because it is strict about only accepting
+    # unicode input.
+    from cStringIO import StringIO as BytesIO
+except ImportError:
+    from io import BytesIO
 import datetime
 try:
     from email.mime.multipart import MIMEMultipart
@@ -52,8 +58,14 @@ except ImportError:
     from email.MIMENonMultipart import MIMENonMultipart
 
 import time
-import urllib
-import simplejson
+try:
+    from urllib.parse import urlencode
+except ImportError:
+    from urllib import urlencode
+try:
+    import simplejson as json
+except ImportError:
+    import json
 try:
     import xml.etree.cElementTree as ET
 except ImportError:
@@ -63,7 +75,11 @@ except ImportError:
         import elementtree.ElementTree as ET
 from lazr.uri import URI, merge
 
-from iso_strptime import iso_strptime
+from wadllib import (
+    _make_unicode,
+    _string_types,
+    )
+from wadllib.iso_strptime import iso_strptime
 
 NS_MAP = "xmlns:map"
 XML_SCHEMA_NS_URI = 'http://www.w3.org/2001/XMLSchema'
@@ -267,7 +283,7 @@ class Resource(WADLResolvableDefinition):
         """
         super(Resource, self).__init__(application)
         self._url = url
-        if isinstance(resource_type, basestring):
+        if isinstance(resource_type, _string_types):
             # We were passed the URL to a resource type. Look up the
             # type object itself
             self.tag = self.application.get_resource_type(resource_type).tag
@@ -280,7 +296,8 @@ class Resource(WADLResolvableDefinition):
         if representation is not None:
             if media_type == 'application/json':
                 if representation_needs_processing:
-                    self.representation = simplejson.loads(unicode(representation))
+                    self.representation = json.loads(
+                        _make_unicode(representation))
                 else:
                     self.representation = representation
             else:
@@ -388,6 +405,19 @@ class Resource(WADLResolvableDefinition):
                     return method
         return None
 
+    def parameters(self, media_type=None):
+        """A list of this resource's parameters.
+
+        :param media_type: Media type of the representation definition
+            whose parameters are being named. Must be present unless
+            this resource is bound to a representation.
+
+        :raise NoBoundRepresentationError: If this resource is not
+            bound to a representation and media_type was not provided.
+        """
+        return self._find_representation_definition(
+            media_type).params(self)
+
     def parameter_names(self, media_type=None):
         """A list naming this resource's parameters.
 
@@ -522,10 +552,13 @@ class Resource(WADLResolvableDefinition):
 
         :return: A RepresentationDefinition
         """
-        if media_type is not None:
-            definition = self.get_representation_definition(media_type)
-        elif self.representation is not None:
+        if self.representation is not None:
+            # We know that when this object was created, a
+            # representation definition was either looked up, or
+            # directly passed in along with the representation.
             definition = self.representation_definition.resolve_definition()
+        elif media_type is not None:
+            definition = self.get_representation_definition(media_type)
         else:
             raise NoBoundRepresentationError(
                 "Resource is not bound to any representation, and no media "
@@ -699,7 +732,7 @@ class RequestDefinition(WADLBase, HasParametersMixin):
                 append = '&'
             else:
                 append = '?'
-            url += append + urllib.urlencode(validated_values)
+            url += append + urlencode(validated_values)
         return url
 
 
@@ -801,7 +834,7 @@ class RepresentationDefinition(WADLResolvableDefinition, HasParametersMixin):
             params, param_values, **kw_param_values)
         media_type = self.media_type
         if media_type == 'application/x-www-form-urlencoded':
-            doc = urllib.urlencode(validated_values)
+            doc = urlencode(validated_values)
         elif media_type == 'multipart/form-data':
             outer = MIMEMultipart()
             outer.set_type('multipart/form-data')
@@ -822,12 +855,13 @@ class RepresentationDefinition(WADLResolvableDefinition, HasParametersMixin):
                     outer.attach(inner)
             doc = str(outer)
             # Chop off the 'From' line, which only makes sense in an
-            # email.
-            doc = doc[doc.find('\n')+1:]
+            # email.  (Python 3 does not include it.)
+            if doc.startswith('From '):
+                doc = doc[doc.find('\n')+1:]
             media_type = (outer.get_content_type() +
                           '; boundary="%s"' % outer.get_boundary())
         elif media_type == 'application/json':
-            doc = simplejson.dumps(validated_values)
+            doc = json.dumps(validated_values)
         else:
             raise ValueError("Unsupported media type: '%s'" % media_type)
         return media_type, doc
@@ -910,9 +944,23 @@ class Parameter(WADLBase):
         return [Option(self, option_tag)
                 for option_tag  in self.tag.findall(wadl_xpath('option'))]
 
+    @property
+    def link(self):
+        """Get the link to another resource.
+
+        The link may be examined and, if its type is of a known WADL
+        description, it may be followed.
+
+        :return: A Link object, or None.
+        """
+        link_tag = self.tag.find(wadl_xpath('link'))
+        if link_tag is None:
+            return None
+        return Link(self, link_tag)
+
     @property
     def linked_resource(self):
-        """Find the type of resource linked to by this parameter.
+        """Follow a link from this parameter to a new resource.
 
         This only works for parameters whose WADL definition includes a
         <link> tag that points to a known WADL description.
@@ -920,10 +968,10 @@ class Parameter(WADLBase):
         :return: A Resource object for the resource at the other end
         of the link.
         """
-        link_tag = self.tag.find(wadl_xpath('link'))
-        if link_tag is None:
+        link = self.link
+        if link is None:
             raise ValueError("This parameter isn't a link to anything.")
-        return Link(self, link_tag).resolve_definition()
+        return link.follow
 
 class Option(WADLBase):
     """One of a set of possible values for a parameter."""
@@ -946,7 +994,7 @@ class Link(WADLResolvableDefinition):
     """A link from one resource to another.
 
     Calling resolve_definition() on a Link will give you a Resource for the
-    type of resource linked to.
+    type of resource linked to. An alias for this is 'follow'.
     """
 
     def __init__(self, parameter, link_tag):
@@ -959,6 +1007,28 @@ class Link(WADLResolvableDefinition):
         self.parameter = parameter
         self.tag = link_tag
 
+    @property
+    def follow(self):
+        """Follow the link to another Resource."""
+        if not self.can_follow:
+            raise WADLError("Cannot follow a link when the target has no "
+                            "WADL description. Try using a general HTTP "
+                            "client instead.")
+        return self.resolve_definition()
+
+    @property
+    def can_follow(self):
+        """Can this link be followed within wadllib?
+
+        wadllib can follow a link if it points to a resource that has
+        a WADL definition.
+        """
+        try:
+            definition_url = self._get_definition_url()
+        except WADLError:
+            return False
+        return True
+
     def _definition_factory(self, id):
         """Turn a resource type ID into a ResourceType."""
         return Resource(
@@ -996,8 +1066,9 @@ class Application(WADLBase):
         """
         self.markup_url = markup_url
         if hasattr(markup, 'read'):
-            markup = markup.read()
-        self.doc = self._from_string(markup)
+            self.doc = self._from_stream(markup)
+        else:
+            self.doc = self._from_string(markup)
         self.resources = self.doc.find(wadl_xpath('resources'))
         self.resource_base = self.resources.attrib.get('base')
         self.representation_definitions = {}
@@ -1012,7 +1083,7 @@ class Application(WADLBase):
             id = resource_type.attrib['id']
             self.resource_types[id] = ResourceType(resource_type)
 
-    def _from_string(self, markup):
+    def _from_stream(self, stream):
         """Turns markup into a document.
 
         Just a wrapper around ElementTree which keeps track of namespaces.
@@ -1021,7 +1092,7 @@ class Application(WADLBase):
         root = None
         ns_map = []
 
-        for event, elem in ET.iterparse(StringIO(markup), events):
+        for event, elem in ET.iterparse(stream, events):
             if event == "start-ns":
                 ns_map.append(elem)
             elif event == "end-ns":
@@ -1032,6 +1103,14 @@ class Application(WADLBase):
                 elem.set(NS_MAP, dict(ns_map))
         return ET.ElementTree(root)
 
+    def _from_string(self, markup):
+        """Turns markup into a document.
+
+        Just a wrapper around ElementTree which keeps track of namespaces.
+        """
+        # We're using the ET fromstring/tostring to help make sure we maintain
+        # py2/3 compatiblity and keeping things in Unicode.
+        return self._from_stream(BytesIO(ET.tostring(ET.fromstring(markup))))
 
     def get_resource_type(self, resource_type_url):
         """Retrieve a resource type by the URL of its description."""
